# Copyright: 2017, CCX Technologies
"""D-Bus Method"""

import functools
import inspect

from .. import sdbus


class Method:
    """Provides an interface between a D-Bus and a Python Method or Function.

    Though this class can be used on its own it is intended to be used with
    the method decorator applied to methods on a class that is a subclass of
    the main Object type.

    Args:
        callback (function): function that will be called when
            D-Bus method is called, it should use new style type
            annotations to define the types of all arguments,
            and the return value, ie. function(x: int, y: str) -> bool:
            if no type is defined a D-Bus Variant will be used
        name (str): optional, name of the method advertised on the D-Bus
            if not set then will use the callback name
        depreciated (bool): optional, if true object is labelled
            as depreciated in the introspect XML data
        hidden (bool): optional, if true object won't be added
            to the introspect XML data
        unprivileged (bool): optional, indicates that this method
            may have to ask the user for authentication before
            executing, which may take a little while
        camel_convert (bool): optional, D-Bus method and property
            names are typically defined in Camel Case, but Python
            methods and arguments are typically defined in Snake
            Case, if this is set the cases will be automatically
            converted between the two
        dont_block (bool): optional, if true the method call will not
            block on the D-Bus, **a value will never be returned**

    Raises:
        BusError: if an error occurs during initialization
    """

    def __init__(
        self,
        callback,
        name=None,
        depreciated=False,
        hidden=False,
        unprivileged=False,
        camel_convert=True,
        dont_block=False,
    ):
        if not name:
            name = callback.__name__

        self.callback = callback
        """Function called when this method is called."""

        if camel_convert:
            self.dbus_name = sdbus.snake_to_camel(name)
            """Method name advertised on the D-Bus."""
        else:
            self.dbus_name = name

        self.depreciated = depreciated
        self.hidden = hidden
        self.unprivileged = unprivileged

        self.arg_signature = ''
        sig = inspect.signature(callback)
        for param in sig.parameters.values():
            if param.annotation != inspect.Parameter.empty:
                self.arg_signature += sdbus.dbus_signature(param.annotation)
            else:
                self.arg_signature += sdbus.variant_signature()

        if dont_block:
            self.return_signature = ''
        elif sig.return_annotation != inspect.Parameter.empty:
            self.return_signature = sdbus.dbus_signature(sig.return_annotation)
        else:
            self.return_signature = sdbus.variant_signature()

    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)

    def vt(self, instance=None):
        """Interface to sd-bus library"""
        if instance:
            arg_signature = self.arg_signature[1:]
            callback = functools.partial(self.callback, instance)
        else:
            arg_signature = self.arg_signature
            callback = self.callback

        return sdbus.Method(
            self.dbus_name, callback, arg_signature, self.return_signature,
            self.depreciated, self.hidden, self.unprivileged
        )


def method(
    name=None,
    depreciated=False,
    hidden=False,
    unprivileged=False,
    camel_convert=True
):
    """D-Bus Method Decorator.

    Note:
        The decorated method should use new style type annotations
        to define the types of all arguments, and the return value,
        ie. function(x: int, y: str) -> bool:
        if no type is defined a D-Bus Variant will be used

    Args:
        name (str): optional, if set this name will be used on
            the D-Bus, instead of the decorated function's name
        depreciated (bool): optional, if true object is labelled
            as depreciated in the introspect XML data
        hidden (bool): optional, if true object won't be added
            to the introspect XML data
        unprivileged (bool): optional, indicates that this method
            may have to ask the user for authentication before
            executing, which may take a little while
        camel_convert (bool): optional, D-Bus method and property
            names are typically defined in Camel Case, but Python
            methods and arguments are typically defined in Snake
            Case, if this is set the cases will be automatically
            converted between the two

    Returns:
        Instantiated Method, which can be used to replace a method
        or function.
    """

    def wrapper(function):
        return Method(
            function, name, depreciated, hidden, unprivileged, camel_convert
        )

    return wrapper



# Copyright: 2017, CCX Technologies
"""D-Bus Service"""

from .. import sdbus
from .. import exceptions
import asyncio


class Service:
    """Serves objects onto a D-Bus, runs within an asyncio loop.

    This is a class can be used to create a server that attaches the to
    D-Bus and serves objects. To process requests it must be run within an
    asyncio loop.

    Objects are added by initializing them with an instantiated service.
    More than one object can be attached to a service as long as they
    don't have the same path.

    Args:
        name (str): name to be used on the D-Bus, ie. org.test
        loop: optional, asyncio loop to attach to, if not
            defined will use asyncio.get_event_loop()
        bus (str): optional, Bus to attach to either 'system', or 'session'
        replace_existing (bool): optional, if true and there is a
            D-Bus service with the same name, try to replace it
        allow_replacement (bool): optional, if true and another service
            tries to steal our name, let it
        name_queue (bool): optional, if our name is in use by another
            service put ourselves in the D-Bus name queue, we will get
            name once the existing service relinquishes it

    Raises:
        BusError: if an error occurs during initialization

    """

    def __init__(
        self,
        name,
        loop=None,
        bus='system',
        replace_existing=False,
        allow_replacement=False,
        name_queue=False
    ):
        self.sdbus = sdbus.Service(
            name, loop, bus, replace_existing, allow_replacement, name_queue
        )
        """Interface to sd-bus library"""

    def is_running(self):
        return self.sdbus.is_running()

    def get_loop(self):
        return self.sdbus.get_loop()



# Copyright: 2017, CCX Technologies
"""D-Bus Object"""

from .. import sdbus


class Object:
    """Provides an interface between a D-Bus and a Python Object.

    Though this class can be used on its own it is intended to be inherited
    by another object which uses decorators to define methods, properties,
    and signals which are exposed on the D-Bus.

    This object must be added to a service, which provides the D-Bus interface.

    It also provides a Context Manager so that multiple properties can be set,
    or one property can be set multiple times, with a single D-Bus changed signal
    being sent for all property updates once the context is closed.

    Args:
        service (adbus.server.Service): service to connect to
        path (str): path to create on the service
            ie. /com/awesome/Settings1
        interface (str): interface label to use for all of this
            objects methods and properties, ie. com.awesome.settings
        vtable (list): optional, list of signals, methods, and
            properties that will be added to the D-Bus Object,
            this list is in addition to the methods, properties, and
            signals which are added to this object using the provided
            decorators and descriptors
        depreciated (bool): optional, if true object is labelled
            as depreciated in the introspect XML data
        hidden (bool): optional, if true object won't be added
            to the introspect XML data
        manager (bool): optional, if True add a device manager to this object,
            as defined by the D-Bus Spec from freedesktop.org
        changed_callback: optional, callback called with a list of changed
            properties, single argument is a list of property names

    Raises:
        BusError: if an error occurs during initialization

    """

    def __init__(
        self,
        service,
        path,
        interface,
        vtable=[],
        depreciated=False,
        hidden=False,
        manager=False,
        changed_callback=None,
    ):
        self._defer_properties = False
        self._deferred_properties = {}

        self.service = service

        self.vtable = [x.vt() for x in vtable]
        """List of all D-Bus Methods, Properties, and Signals."""

        self.vtable += [
            v.vt(self)
            for v in type(self).__dict__.values() if hasattr(v, 'vt')
        ]

        self.sdbus = sdbus.Object(
            service.sdbus, path, interface, self.vtable, depreciated, hidden
        )
        """Interface to sd-bus library."""

        if manager:
            self.manager = sdbus.Manager(service.sdbus, path)

        self.changed_callback = changed_callback

    def emit_property_changed(self, py_name, dbus_name):
        if self._defer_properties:
            self._deferred_properties[dbus_name.encode()] = py_name

        else:
            if self.service.is_running():
                self.sdbus.emit_properties_changed([dbus_name.encode()])

            if self.changed_callback:
                self.service.get_loop().run_in_executor(
                    None, self.changed_callback, [py_name]
                )

    def defer_signals(self, enable):
        if enable:
            self._defer_properties = True

        elif self._defer_properties:
            self._defer_properties = False

            if self._deferred_properties:
                if self.service.is_running():
                    self.sdbus.emit_properties_changed(
                        list(self._deferred_properties.keys())
                    )

            if self.changed_callback:
                self.service.get_loop().run_in_executor(
                    None, self.changed_callback,
                    list(self._deferred_properties.values())
                )

            self._deferred_properties = {}

    def __enter__(self):
        self.defer_signals(True)
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.defer_signals(False)



#!/usr/bin/env python3
# Copyright: 2017, CCX Technologies
"""distutils build and configuration script for python-adbus."""

from distutils.core import setup
from distutils.core import Extension
from Cython.Build import cythonize
import sys

if "--cythonize" in sys.argv:
    sdbus = cythonize(
        [Extension("adbus.sdbus", ["adbus/sdbus.pyx"], libraries=["systemd"])]
    )
else:
    sdbus = [Extension("adbus.sdbus", ["adbus/sdbus.c"], libraries=["systemd"])]

def build():
    """Build Package."""

    setup(
        name='adbus',
        version='0.1.1',
        description='asyncio based dbus interface',
        license='MIT',
        author='CCX Technologies',
        author_email='charles@ccxtechnologies.com',
        url='http://github.com/ccxtechnologies/python-adbus',
        download_url='https://github.com/ccxtechnologies/python-adbus/archive/v0.1.1.tar.gz',
        platforms=['linux'],
        provides=['adbus'],
        packages=['adbus', 'adbus.server'],
        ext_modules=sdbus
    )

if __name__ == "__main__":
    build()
